#include <assert.h>
#include "varset.h"
#include "genkill.h"

//********************************************************************
// GenKillNode
//********************************************************************

GenKillNode::GenKillNode(VarSet k, VarSet g, unsigned int c)
  : kill(Diff(k,g)), gen(g), count(c) {}
#if 0
	  std::cerr << "GenKillNode(" << k << ", " << g << ")\n\t";
	  std::cerr << kill << "\n\t";
	  std::cerr << gen << std::endl;
  }
#endif

GenKillNode::~GenKillNode() {} 

GenKillNode * GenKillNode::one()
{
	return &idnode;
}

GenKillNode * GenKillNode::zero()
{
	return &top;
}

GenKillNode * GenKillNode::extend( GenKillNode * y ) const
{
	VarSet temp_k;
	VarSet temp_g;

	if( equal(GenKillNode::zero()) ) {
		return GenKillNode::zero();
	}
	else if( y->equal(GenKillNode::zero()) ) {
		return GenKillNode::zero();
	}
	temp_k = Union( kill, y->kill );
	temp_g = Union( Diff(gen,y->kill),y->gen);
	return makeGenKillNode( temp_k,temp_g );

} /* end of Extender **/

GenKillNode * GenKillNode::combine( GenKillNode * y ) const
{
	VarSet temp_k, temp_g;

	temp_k = Intersect( kill, y->kill );
	temp_g = Union( gen, y->gen );

	return makeGenKillNode( temp_k,temp_g );
} /** end of Combiner **/

GenKillNode * GenKillNode::quasiOne() const
{
    return one();
}

bool GenKillNode::equal( GenKillNode * y ) const
{
	return Eq(kill,y->kill) && Eq(gen,y->gen);
}

/*
 * diff( GenKillNode * y)
 *
 * Return the difference between x (this) and y.
 *
 * The return value r has two properties:
 *  o r >= x
 *    (i.e., Eq(Combiner(x,r), x) == true)
 *  o y combine r = y combine a
 *    (i.e., Eq(Combiner(y,r), Combiner(y,a)) == true)
 */
GenKillNode * GenKillNode::diff( GenKillNode * y ) const
{
  VarSet temp_k, temp_g;

  temp_k = Diff(UniverseVarSet,Diff(y->kill,kill));
  temp_g = Diff(gen,y->gen);

  return makeGenKillNode(temp_k, temp_g);
}

std::ostream& GenKillNode::print( std::ostream& o ) const
{
	return (o << "<\\S.(S - {" << kill << "}) U {" << gen << "}>");
}

// static vars top,bottom
GenKillNode GenKillNode::top(EmptyVarSet,EmptyVarSet,1);
GenKillNode GenKillNode::bottom(EmptyVarSet,EmptyVarSet,1);
GenKillNode GenKillNode::idnode(EmptyVarSet,EmptyVarSet,1);

// static private method
GenKillNode * GenKillNode::makeGenKillNode(
		VarSet& k,
		VarSet& g )
{
	if (Eq(k,EmptyVarSet) && Eq(g,UniverseVarSet)) {
		return &bottom;
	}
	else if (Eq(k,EmptyVarSet) && Eq(g,EmptyVarSet)) {
		return GenKillNode::one();
	}
	else if (Eq(k,UniverseVarSet) && Eq(g,EmptyVarSet)) {
		return GenKillNode::zero();
	}
	else {
		return new GenKillNode( k, g );
	}
}

void GenKillNode::init()
{
  top.kill = UniverseVarSet;
  bottom.gen = UniverseVarSet;
}

std::ostream& GenKillNode::print_static_nodes( std::ostream& o )
{
	top.print( (std::cout << "TopNode\t=\t") )<< std::endl;
	bottom.print( (std::cout << "BottomNode\t=\t") ) << std::endl;
	idnode.print( (std::cout << "IdNode\t=\t") ) << std::endl;
	return o;
}
